class SnowfakeryPlugin:
    """Base class for all plugins.

    Currently plugins can declare new functions. Over time new superpowers
    for plugins will be introduced.

    Plugins generally look like this:

    class MyPlugin:
        class Functions: # must be named `Functions`.
            def func1(self, arg1, arg2, arg3=default):
                return something()

    This function would be invoked through the name `MyPlugin.func1()`

    Despite the name, the Functions namespace could also include
    constants or other namespaces, if that is useful to you.

    A RuntimeContext object is supplied on `self.context` with methods
    context.field_vars() and context.context_vars().

    context.field_vars() are the same field variables that can be used in
    templates.

    context.context_vars() is a mutable mapping contaiing
    values that are available to this object template and that of
    all children.

    context_vars are unique to each plugin and plugins do not see
    (or corrupt, or overwrite) context_vars for another plugin.

    Plugins can also keep internal state for global data, like any other
    Python object.
    """

    def __init__(self, interpreter):
        self.interpreter = interpreter
        self.context = PluginContext(self)

    def custom_functions(self, *args, **kwargs):
        """Instantiate, contextualize and return a function library

        Default behaviour is to return self.Function()."""
        functions = self.Functions()
        functions.context = self.context
        return functions


class PluginContext:
    "Exposes certain stable internals to plugins"

    def __init__(self, plugin):
        self.plugin = plugin
        self.interpreter = plugin.interpreter

    def field_vars(self):
        return self.interpreter.current_context.field_vars()

    def context_vars(self):
        return self.interpreter.current_context.context_vars(
            self.plugin.__class__.__name__
        )



from collections import defaultdict
from functools import partial
from datetime import date
from contextlib import contextmanager

from typing import Optional, Dict, List, Sequence, Mapping, NamedTuple

from faker.utils.datetime_safe import date as faker_date
import jinja2
import yaml

from .utils.template_utils import FakerTemplateLibrary

from .template_funcs import template_funcs
from .data_gen_exceptions import DataGenSyntaxError, DataGenNameError
import snowfakery  # noqa

OutputStream = "snowfakery.output_streams.OutputStream"


# Runtime objects and algorithms used during the generation of rows.


class StoppingCriteria(NamedTuple):
    """When have we iterated over the Snowfakery script enough times?"""

    tablename: str
    count: int


class FinishedChecker:
    """Checks whether the Stopping Criteria have been met"""

    stopping_criteria: StoppingCriteria

    def __init__(self, start_ids, stopping_criteria):
        self.start_ids = start_ids
        self.stopping_criteria = stopping_criteria
        self.target_progress_id = 0

    def check_if_finished(self, id_manager):
        "Check whether we've finished making as many objects as we promised"
        # if nobody told us how much to make, finish after first run
        if not self.stopping_criteria:
            return True

        target_table, count = self.stopping_criteria

        # Snowfakery processes can be restarted. We would need
        # to keep track of where we restarted to know whether
        # we are truly finished
        if self.start_ids:
            start = self.start_ids.get(target_table, 1)
        else:
            start = 1
        target_id = start + count - 1

        last_used_id = id_manager[target_table]

        self.ensure_progress_was_made(last_used_id, target_id)

        return last_used_id >= target_id

    def ensure_progress_was_made(self, last_used_id, target_id):
        """Check that we are actually making progress towards our goal"""
        if last_used_id == self.target_progress_id:
            raise RuntimeError(
                f"{self.stopping_criteria.tablename} max ID was {self.target_progress_id} "
                f"before evaluating factory and is {last_used_id} after. "
                f"At this rate we will never hit our target of {target_id}!"
            )

        self.target_progress_id = last_used_id


class IdManager(yaml.YAMLObject):
    """Keep track of the most recent ID per Object type"""

    yaml_loader = yaml.SafeLoader
    yaml_dumper = yaml.SafeDumper
    yaml_tag = "!snowfakery_ids"

    def __init__(self, start_ids: Optional[Dict[str, int]] = None):
        start_ids = start_ids or {}
        self.last_used_ids = defaultdict(lambda: 0)
        for name, value in start_ids.items():
            self.last_used_ids[name] = value - 1

    def generate_id(self, table_name: str) -> int:
        self.last_used_ids[table_name] += 1
        return self.last_used_ids[table_name]

    def __getitem__(self, table_name: str) -> int:
        return self.last_used_ids[table_name]

    def __setstate__(self, state):
        self.last_used_ids = defaultdict(lambda: 0, state["last_used_ids"])


yaml.SafeDumper.add_representer(defaultdict, yaml.SafeDumper.represent_dict)


class Dependency(NamedTuple):
    yaml_loader = yaml.SafeLoader
    yaml_dumper = yaml.SafeDumper
    yaml_tag = "!snowfakery_dependency"

    table_name_from: str
    table_name_to: str
    field_name: str


yaml.SafeDumper.add_representer(
    Dependency, lambda representer, obj: representer.represent_list(obj)
)


yaml.SafeDumper.add_representer(
    faker_date, yaml.representer.SafeRepresenter.represent_date
)


class Globals(yaml.YAMLObject):
    """Globally named objects and other aspects of global scope

     This object is designed to be persisted to allow long-running
     Snowfakery executions to stop and restart. Other Interpreter internals
     do not persist. For example, it isn't possible to persist a database
     handle in an output_stream."""

    yaml_loader = yaml.SafeLoader
    yaml_dumper = yaml.SafeDumper
    yaml_tag = "!snowfakery_globals"

    def __init__(self, start_ids: Optional[Dict[str, id]] = None, today: date = None):
        self.named_objects = {}
        self.id_manager = IdManager(start_ids)
        self.last_seen_obj_of_type = {}
        self.intertable_dependencies = set()
        self.today = today or date.today()

    def register_object(self, obj, nickname: str = None):
        """Register an object for lookup by object type and (optionally) Nickname"""
        if nickname:
            self.named_objects[nickname] = obj
        self.last_seen_obj_of_type[obj._tablename] = obj

    @property
    def object_names(self):
        """The globally named objects"""
        return {**self.named_objects, **self.last_seen_obj_of_type}

    def register_intertable_reference(
        self, table_name_from: str, table_name_to: str, fieldname: str
    ):
        self.intertable_dependencies.add(
            Dependency(table_name_from, table_name_to, fieldname)
        )

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["intertable_dependencies"]
        return state

    def __setstate__(self, state):
        self.last_seen_obj_of_type = {}
        for slot, value in state.items():
            setattr(self, slot, value)
        self.intertable_dependencies = set()


class JinjaTemplateEvaluatorFactory:
    def __init__(self):
        self.template_compiler = jinja2.Environment(
            block_start_string="<%",
            block_end_string="%>",
            variable_start_string="<<",
            variable_end_string=">>",
        )

    def get_evaluator(self, definition: str):
        assert isinstance(definition, str), definition
        if "<<" in definition or "<%" in definition:
            try:
                template = self.template_compiler.from_string(definition)
                return lambda context: template.render(context.field_vars())
            except jinja2.exceptions.TemplateSyntaxError as e:
                raise DataGenSyntaxError(str(e), None, None) from e
        else:
            return lambda context: definition


class Interpreter:
    """Snowfakery runtime interpreter state."""

    current_context: "RuntimeContext" = None

    def __init__(
        self,
        output_stream: OutputStream,
        globals: Globals = None,
        options: Mapping = None,
        snowfakery_plugins: Optional[Mapping[str, callable]] = None,
        stopping_criteria: Optional[StoppingCriteria] = None,
        start_ids: Optional[Mapping[str, int]] = None,
        faker_providers: Sequence[object] = (),
    ):
        self.output_stream = output_stream
        self.options = options or {}
        snowfakery_plugins = snowfakery_plugins or {}
        # todo: some caching of these could speed things up a bit
        self.plugin_function_libraries = {
            name: plugin(self).custom_functions()
            for name, plugin in snowfakery_plugins.items()
        }
        self.finished_checker = FinishedChecker(start_ids, stopping_criteria)
        self.faker_template_library = FakerTemplateLibrary(faker_providers)
        self.globals = globals or Globals()


class RuntimeContext:
    """Local "stack frame" type object. RuntimeContexts live on the Python stack.

    There are many proxy methods for other objects which helps keep the
    internals of the runtime hidden from outside code to make maintenance
    easier. From the point of view of other modules this is "the interpreter"
    but internally its mostly just proxying to other classes."""

    obj: Optional["ObjectRow"] = None
    template_evaluator_factory = JinjaTemplateEvaluatorFactory()

    def __init__(
        self,
        interpreter: Interpreter,
        current_table_name: Optional[str] = None,
        parent_context: "RuntimeContext" = None,
    ):
        self.current_table_name = current_table_name
        self.interpreter = interpreter
        self.interpreter.current_context = self
        self.parent = parent_context
        if self.parent:
            self._context_vars = {
                **self.parent._context_vars
            }  # implements variable inheritance
        else:
            self._context_vars = {}

    def check_if_finished(self):
        "Have we iterated over the script enough times?"
        return self.interpreter.finished_checker.check_if_finished(
            self.interpreter.globals.id_manager
        )

    def generate_id(self):
        "Generate a unique ID for this object"
        return self.interpreter.globals.id_manager.generate_id(self.current_table_name)

    def register_object(self, obj, name=None):
        "Keep track of this object in case its children refer to it."
        self.obj = obj
        self.interpreter.globals.register_object(obj, name)

    def register_intertable_reference(self, table_name_from, table_name_to, fieldname):
        "Keep track of a dependency between two tables. e.g. for mapping file generation"
        self.interpreter.globals.register_intertable_reference(
            table_name_from, table_name_to, fieldname
        )

    @contextmanager
    def child_context(self, tablename):
        "Create a nested RuntimeContext (analogous to a 'stack frame')."
        jr = self.__class__(
            current_table_name=tablename,
            interpreter=self.interpreter,
            parent_context=self,
        )
        # jr will register itself with the interpreter
        try:
            yield jr
        finally:
            # Goodbye junior, its been nice
            # Hope you find your paradise
            self.interpreter.current_context = self

    @property
    def output_stream(self):
        return self.interpreter.output_stream

    def get_evaluator(self, definition: str):
        return self.template_evaluator_factory.get_evaluator(definition)

    @property
    def evaluation_namespace(self):
        return EvaluationNamespace(self)

    def executable_blocks(self):
        return self.evaluation_namespace.executable_blocks()

    def field_vars(self):
        return self.evaluation_namespace.field_vars()

    def context_vars(self, plugin_namespace):
        return self._context_vars.setdefault(plugin_namespace, {})


# NamedTuple because it is immutable, efficient and auto-generates init
class EvaluationNamespace(NamedTuple):
    """Supplies names for evaluation of YAML trees and Jinja expressions."""

    runtime_context: RuntimeContext

    def simple_field_vars(self):
        "Variables that can be inserted into templates"
        # obj=None in some contexts, e.g. evaluating count
        obj = self.runtime_context.obj
        interpreter = self.runtime_context.interpreter
        return {
            "id": obj.id if obj else None,
            "count": obj.id if obj else None,
            "this": obj,
            "today": interpreter.globals.today,
            "fake": interpreter.faker_template_library,
            "fake_i18n": lambda locale: FakerTemplateLibrary(locale),
            **interpreter.options,
            **interpreter.globals.object_names,
            **(obj._values if obj else {}),
            **interpreter.plugin_function_libraries,
        }

    # todo: should be replaced with the plugin architecture
    def field_funcs(self):
        "Injects context into functions from template_funcs module."

        def curry(func):
            rc = partial(func, self.runtime_context)
            if hasattr(func, "lazy"):
                rc.lazy = func.lazy
            return rc

        funcs = {name: curry(func) for name, func in template_funcs.items()}
        return {**funcs}

    def executable_blocks(self):
        "Return mapping of functions that can be used in YAML block functions"
        return {**self.field_funcs(), "fake": self.fake}

    # todo: remove this special case
    def fake(self, name):
        return str(
            getattr(self.runtime_context.interpreter.faker_template_library, name)
        )

    def field_vars(self):
        return {**self.simple_field_vars(), **self.field_funcs()}


def evaluate_function(func, args: Sequence, kwargs: Mapping, context):
    if not hasattr(func, "lazy"):
        args = [arg.render(context) if hasattr(arg, "render") else arg for arg in args]
        kwargs = {
            name: arg.render(context) if hasattr(arg, "render") else arg
            for name, arg in kwargs.items()
        }
    value = func(*args, **kwargs)
    return value


class ObjectRow(yaml.YAMLObject):
    """Represents a single row

    Uses __getattr__ so that the template evaluator can use dot-notation."""

    yaml_loader = yaml.SafeLoader
    yaml_dumper = yaml.SafeDumper
    yaml_tag = "!snowfakery_objectrow"

    __slots__ = ["_tablename", "_values"]

    def __init__(self, tablename, values=()):
        self._tablename = tablename
        self._values = values

    def __getattr__(self, name):
        try:
            return self._values[name]
        except KeyError:
            raise AttributeError(name)

    def __str__(self):
        return str(self.id)

    def __repr__(self):
        return f"<ObjectRow {self._tablename} {self.id}>"

    def __getstate__(self):
        return {"_tablename": self._tablename, "_values": self._values}

    def __setstate__(self, state):
        for slot, value in state.items():
            setattr(self, slot, value)

    @property
    def _name(self):
        return self._values.get("name")


def output_batches(
    output_stream,
    factories: List,
    options: Dict,
    stopping_criteria: Optional[StoppingCriteria] = None,
    continuation_data: Globals = None,
    tables: Mapping[str, int] = None,
    snowfakery_plugins: Mapping[str, snowfakery.SnowfakeryPlugin] = None,
    faker_providers: List[object] = None,
) -> Globals:
    """Generate 'count' batches to 'output_stream' """
    # check the stopping_criteria against the factories available
    if stopping_criteria:
        stop_table_name = stopping_criteria.tablename
        if stop_table_name not in tables:
            raise DataGenNameError(
                f"No template creating {stop_table_name}", None, None
            )

    if continuation_data:
        globals = continuation_data
        start_ids = {
            name: val + 1 for name, val in globals.id_manager.last_used_ids.items()
        }
    else:
        globals = Globals()
        # at one point start-ids were passed from the command line
        # but for simplicity this feature has been removed and they can only
        # be inferred from the continuation_file.
        start_ids = {}

    interpreter = Interpreter(
        output_stream=output_stream,
        globals=globals,
        options=options,
        snowfakery_plugins=snowfakery_plugins,
        stopping_criteria=stopping_criteria,
        start_ids=start_ids,
        faker_providers=faker_providers,
    )

    runtimecontext = RuntimeContext(interpreter=interpreter)
    continuing = bool(continuation_data)
    loop_over_factories(factories, runtimecontext, output_stream, continuing)
    return interpreter.globals


def loop_over_factories(factories, runtimecontext, output_stream, continuing):
    finished = False
    while not finished:
        for factory in factories:
            should_skip = factory.just_once and continuing is True
            if not should_skip:
                factory.generate_rows(output_stream, runtimecontext)
        finished = runtimecontext.check_if_finished()

        continuing = True



import random
from functools import lru_cache
from datetime import date, datetime
import dateutil.parser
from ast import literal_eval

from typing import Callable, Any, Optional, Union, List, Tuple

from faker import Faker

from .data_gen_exceptions import DataGenError

import snowfakery.data_generator_runtime  # noqa

RuntimeContext = "snowfakery.data_generator_runtime.RuntimeContext"
FieldDefinition = "snowfakery.data_generator_runtime_dom.FieldDefinition"
ObjectRow = "snowfakery.data_generator_runtime.ObjectRow"

fake = Faker()

# It might make more sense to use context vars for context handling when
# Python 3.6 is out of the support matrix.


def lazy(func: Any) -> Callable:
    """A lazy function is one that expects its arguments to be unparsed"""
    func.lazy = True
    return func


def random_number(context: RuntimeContext, min: int, max: int) -> int:
    """Pick a random number between min and max like Python's randint."""
    return random.randint(min, max)


def parse_weight_str(context: RuntimeContext, weight_value) -> int:
    """For constructs like:

    - choice:
        probability: 60%
        pick: Closed Won

    Render and convert the 60% to just 60.
    """
    weight_str = weight_value.render(context)
    if isinstance(weight_str, str):
        weight_str = weight_str.rstrip("%")
    return int(weight_str)


def weighted_choice(choices: List[Tuple[int, object]]):
    """Selects from choices based on their weights"""
    weights = [weight for weight, value in choices]
    options = [value for weight, value in choices]
    return random.choices(options, weights, k=1)[0]


@lazy
def random_choice(context: RuntimeContext, *choices):
    """Template helper for random choices.

    Supports structures like this:

    random_choice:
        - a
        - b
        - <<c>>

    Or like this:

    random_choice:
        - choice:
            pick: A
            probability: 50%
        - choice:
            pick: A
            probability: 50%

    Probabilities are really just weights and don't need to
    add up to 100.

    Pick-items can have arbitrary internal complexity.

    Pick-items are lazily evaluated.
    """
    if not choices:
        raise ValueError("No choices supplied!")

    if getattr(choices[0], "function_name", None) == "choice":
        choices = [choice.render(context) for choice in choices]
        rc = weighted_choice(choices)
    else:
        rc = random.choice(choices)
    if hasattr(rc, "render"):
        rc = rc.render(context)
    return rc


@lazy
def choice_wrapper(
    context: RuntimeContext,
    pick,
    probability: FieldDefinition = None,
    when: FieldDefinition = None,
):
    """Supports the choice: sub-items used in `random_choice` or `if`"""
    if probability:
        probability = parse_weight_str(context, probability)
    return probability or when, pick


@lru_cache(maxsize=512)
def parse_date(d: Union[str, datetime, date]) -> Optional[Union[datetime, date]]:
    if isinstance(d, (datetime, date)):
        return d
    try:
        return dateutil.parser.parse(d)
    except dateutil.parser.ParserError:
        pass


def date_(
    context: RuntimeContext,
    *,
    year: Union[str, int],
    month: Union[str, int],
    day: Union[str, int],
):
    """A YAML-embeddable function to construct a date from strings or integers"""
    return date(year, month, day)


def datetime_(
    context: RuntimeContext,
    *,
    year: Union[str, int],
    month: Union[str, int],
    day: Union[str, int],
    hour=0,
    minute=0,
    second=0,
    microsecond=0,
):
    """A YAML-embeddable function to construct a datetime from strings or integers"""
    return datetime(year, month, day, hour, minute, second, microsecond)


def date_between(context: RuntimeContext, start_date, end_date):
    """A YAML-embeddable function to pick a date between two ranges"""
    start_date = parse_date(start_date) or start_date
    end_date = parse_date(end_date) or end_date
    try:
        return fake.date_between(start_date, end_date)
    except ValueError as e:
        if "empty range" not in str(e):
            raise
    # swallow empty range errors per Python conventions


def reference(context: RuntimeContext, x: Union[ObjectRow, str]):
    """YAML-embeddable function to Reference another object."""
    if hasattr(x, "id"):  # reference to an object with an id
        target = x
    elif isinstance(x, str):  # name of an object
        obj = context.field_vars()[x]
        if not getattr(obj, "id"):
            raise DataGenError(f"Reference to incorrect object type {obj}", None, None)
        target = obj
    else:
        raise DataGenError(
            f"Can't get reference to object of type {type(x)}: {x}", None, None
        )

    return target


def render_boolean(context: RuntimeContext, value: FieldDefinition) -> bool:
    val = value.render(context)
    if isinstance(val, str):
        val = literal_eval(val)

    return bool(val)


@lazy
def if_(context: RuntimeContext, *choices: FieldDefinition):
    """Template helper for conditional choices.

    Supports structures like this:

    if:
        - choice:
            when: <<something>>
            pick: A
        - choice:
            when: <<something>>
            pick: B

    Pick-items can have arbitrary internal complexity.

    Pick-items are lazily evaluated.
    """
    if not choices:
        raise ValueError("No choices supplied!")

    choices = [choice.render(context) for choice in choices]
    for when, choice in choices[:-1]:
        if when is None:
            raise SyntaxError(
                "Every choice except the last one should have a when-clause"
            )
    true_choices = (
        choice for when, choice in choices if when and render_boolean(context, when)
    )
    rc = next(true_choices, choices[-1][-1])  # default to last choice
    if hasattr(rc, "render"):
        rc = rc.render(context)
    return rc


template_funcs = {
    "int": lambda context, number: int(number),
    "choice": choice_wrapper,
    "random_number": random_number,
    "random_choice": random_choice,
    "date_between": date_between,
    "reference": reference,
    "date": date_,
    "datetime": datetime_,
    "if": if_,
}



class DataGenError(Exception):
    def __init__(self, message, filename, line_num):
        self.message = message
        self.filename = filename
        self.line_num = line_num
        assert isinstance(filename, (str, type(None)))
        assert isinstance(line_num, (int, type(None)))
        super().__init__(self.message)

    def __str__(self):
        return f"{self.message}\n near {self.filename}:{self.line_num}"


class DataGenSyntaxError(DataGenError):
    pass


class DataGenNameError(DataGenError):
    pass


class DataGenValueError(DataGenError):
    pass


def fix_exception(message, parentobj, e):
    """Add filename and linenumber to an exception if needed"""
    filename, line_num = parentobj.filename, parentobj.line_num
    if isinstance(e, DataGenError):
        if not e.filename:
            e.filename = filename
        if not e.line_num:
            e.line_num = line_num
        raise e
    else:
        raise DataGenError(message, filename, line_num) from e
